/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.GenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ProvidingGenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.start.Opt4JTask;

import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;

/**
 * {@link CreatorModule} that is the entry point and majestro for the create phase of the DI-based
 * DSE process exploration.
 * 
 * @author diewald
 */
public class GuiceCreator extends AbstractModule implements Creator<ProvidingGenotype> {

	/** Set of {@link Genotype}-Types for which instances shall be created. */
	private Set<Class<? extends Genotype>> requiredGenotypes;

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<CreatorModule<?>> enabledCreators;

	/**
	 * Module of genotype provider prototypes. Used to construct the instance providers passed to
	 * the decoding phase.
	 */
	protected EncodingProviderModule<Genotype> protoProviderModule;

	/** {@link Opt4JTask} provider used to obtain the opt4j injector. */
	@Inject
	protected Opt4JDseTaskProvider opt4JProvider;

	/** Child {@link Injector} for creating the required {@link Genotype}s in the creation phase. */
	protected Injector createInjector;

	/** Constructor. */
	public GuiceCreator(Set<Class<? extends Genotype>> requiredGenotypes,
			Collection<CreatorModule<?>> enabledCreators) {
		this.requiredGenotypes = requiredGenotypes;
		this.enabledCreators = enabledCreators;
		// We are "recreating" the prototype provider here (see DseProcessModule -> configure), to
		// avoid "empty" providers flying around since not every enabled genotype might be actually
		// required.
		this.protoProviderModule =
				new GenotypeProviderGenerator().generatePrototypes(requiredGenotypes);
	}

	/** {@inheritDoc} */
	@Override
	public ProvidingGenotype create() {
		createInjector = opt4JProvider.get().get().createChildInjector(this);
		CreateInterceptor interceptor = createInjector.getInstance(CreateInterceptor.class);
		interceptor.setInterceptParameters(createInjector, Genotype.class, GenotypeCache.class);

		GenotypeCache genotypeCache = createInjector.getInstance(GenotypeCache.class);
		genotypeCache.clear();

		for(Class<? extends Genotype> requiredGenotype : requiredGenotypes) {
			createInjector.getInstance(requiredGenotype);
		}

		EncodingProviderModule<Genotype> providerModule =
				new GenotypeProviderGenerator().generateInstanceProviders(protoProviderModule,
						genotypeCache.values());
		// With a composite genotype of composite genotypes and a recursive collect, we could also
		// support arbitrarily deep nested structure, but this not needed atm.
		CompositeGenotype<Class<? extends Genotype>, Genotype> compGenotype =
				new CompositeGenotype<>(genotypeCache);
		return new ProvidingGenotype(providerModule, compGenotype);
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		install(new GenotypeCache());

		for(CreatorModule<?> module : enabledCreators) {
			install(module);
		}
	}
}
