/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static com.google.common.base.Preconditions.checkState;

import java.util.HashMap;
import java.util.Map;

import com.google.common.collect.Maps;
import com.google.inject.Key;
import com.google.inject.Provider;
import com.google.inject.Scope;
import com.google.inject.Scopes;

/**
 * This custom scope can be used to annotate methods or classes with the {@link ThreadScoped}
 * annotation such that the same instance of an object is provided per thread, once it has been
 * constructed in a thread. It is similar to the Session scope.
 * The code is a modified version of the "SimpleScope" example from the Guice wiki page.
 *
 * @author Jesse Wilson
 * @author Fedor Karpelevitch
 * @author Alexander Diewald (diewald@foritss.org)
 */
public class ThreadScope implements Scope {

	/** Internal seed provider. */
	private static final Provider<Object> SEEDED_KEY_PROVIDER = new Provider<Object>() {
		@Override
		public Object get() {
			throw new IllegalStateException("If you got here then it means that"
					+ " your code asked for scoped object which should have been"
					+ " explicitly seeded in this scope by calling"
					+ " SimpleScope.seed(), but was not.");
		}
	};

	/** Map storing instances per thread. */
	private final ThreadLocal<Map<Key<?>, Object>> values = new ThreadLocal<Map<Key<?>, Object>>();

	/** Called (automatically) when entering the scope. */
	public void enter() {
		checkState(values.get() == null, "A scoping block is already in progress");
		values.set(Maps.<Key<?>, Object>newHashMap());
	}

	/** Called (automatically) when leaving the scope. */
	public void exit() {
		checkState(values.get() != null, "No scoping block in progress");
		values.remove();
	}

	/** Adds the given key-value pair to the internal map. */
	public <T> void seed(Key<T> key, T value) {
		Map<Key<?>, Object> scopedObjects = values.get();
		checkState(!scopedObjects.containsKey(key), "A value for the key %s was "
				+ "already seeded in this scope. Old value: %s New value: %s", key,
				scopedObjects.get(key), value);
		scopedObjects.put(key, value);
	}

	/** Adds the given key-value pair to the internal map. */
	public <T> void seed(Class<T> clazz, T value) {
		seed(Key.get(clazz), value);
	}

	/** {@inheritDoc} */
	@Override
	public <T> Provider<T> scope(final Key<T> key, final Provider<T> unscoped) {
		return new Provider<T>() {
			@Override
			public T get() {
				Map<Key<?>, Object> scopedObjects = values.get();
				if(scopedObjects == null) {
					scopedObjects = new HashMap<>();
					values.set(scopedObjects);
				}

				@SuppressWarnings("unchecked") T current = (T)scopedObjects.get(key);
				if(current == null && !scopedObjects.containsKey(key)) {
					current = unscoped.get();

					// don't remember proxies; these exist only to serve circular dependencies
					if(Scopes.isCircularProxy(current)) {
						return current;
					}

					scopedObjects.put(key, current);
				}
				return current;
			}
		};
	}

	/**
	 * Returns a provider that always throws exception complaining that the object
	 * in question must be seeded before it can be injected.
	 *
	 * @return typed provider
	 */
	@SuppressWarnings({"unchecked"})
	public static <T> Provider<T> seededKeyProvider() {
		return (Provider<T>)SEEDED_KEY_PROVIDER;
	}
}
