/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.opt4j.core.Genotype;

import com.google.inject.Provider;
import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * {@link ComposableDecoder} that generates {@link ITaskAdapter}s from the
 * chosen architecture of a safety function.
 * 
 * @author diewald
 */
public class SFGraphDecoder extends DecoderModule<SafeTaskGraphEncoding> implements
		Provider<SafetyFunctionArchEncoding> {

	/**
	 * Decodes the given {@link SafetyFunctionArchEncoding} into a {@link SafeTaskGraphEncoding}
	 * where potential replicas dictated by the safety architectures are present.
	 */
	@Provides
	@Decodes
	public SafeTaskGraphEncoding decode(@Genotyped SafetyFunctionArchEncoding sfaEnc,
			AbstractTaskGraphEncoding atgEnc,
			@Named("OriginalTaskGraphs") Set<TaskGraph> origTaskGraphs) {
		// FIXME: Find a way to pass the original task graph. We might have a circular dependency
		// here.
		SafeTaskGraphEncoding encoding =
				new SafeTaskGraphEncoding(origTaskGraphs, atgEnc.getHyperPeriod());

		for(SafetyFunctionArchEntry sFAEntry : sfaEnc.getAllEntries()) {
			boolean diagPresent = sFAEntry.isDiagnosicUnitPresent();
			if(!diagPresent && containsDiagUnits(sFAEntry, encoding)) {
				removeDiagUnits(sFAEntry, encoding);
			} else if(diagPresent && !containsDiagUnits(sFAEntry, encoding)) {
				addDiagUnits(sFAEntry, encoding);
			}
		}

		return encoding;
	}

	/**
	 * Checks whether the given {@link TaskGraphEncoding} contains diagnostic units from the given
	 * {@link SafetyFunctionArchEntry}.
	 */
	// TODO: This method should be encapsulated within the SFA entry or the adapter to avoid the raw
	// casting here.
	private boolean containsDiagUnits(SafetyFunctionArchEntry sFAEntry,
			SafeTaskGraphEncoding encoding) {
		@SuppressWarnings("unchecked") Collection<ITaskAdapter<?>> sfComps =
				(Collection<ITaskAdapter<?>>)(Collection<?>)sFAEntry.getSafetyFunctionAdapter()
						.getDiagnosticComponents();
		if(!sfComps.isEmpty()) {
			return encoding.containsTask(sfComps.stream().findAny().get());
		}
		return false;
	}

	/**
	 * Adds the diagnostic units from the {@link SafetyFunctionArchEntry} to the
	 * {@link TaskGraphEncoding}.
	 */
	private void addDiagUnits(SafetyFunctionArchEntry sFAEntry, SafeTaskGraphEncoding encoding) {
		@SuppressWarnings("unchecked") Collection<ITaskAdapter<?>> diagComps =
				(Collection<ITaskAdapter<?>>)(Collection<?>)sFAEntry.getSafetyFunctionAdapter()
						.getDiagnosticComponents();
		addTasks(diagComps, encoding);
	}

	/**
	 * Adds the given collection of {@link ITaskAdapter}s from the {@link SafetyFunctionArchEntry}
	 * to the {@link TaskGraphEncoding}.
	 */
	private void addTasks(Collection<ITaskAdapter<?>> tasks, SafeTaskGraphEncoding encoding) {
		// Add the Components.
		tasks.stream().forEach(task -> encoding.addTask(task));

		// Add the corresponding signals.
		for(ITaskAdapter<?> task : tasks) {
			TaskGraph tg = encoding.getOriginalTaskGraph(task);
			encoding.addSignalsOf(task, tg);
		}
	}

	/** Removes the diagnostic units from the {@link TaskGraphEncoding}. */
	private void removeDiagUnits(SafetyFunctionArchEntry sFAEntry, SafeTaskGraphEncoding encoding) {
		@SuppressWarnings("unchecked") Collection<ITaskAdapter<?>> sfComps =
				(Collection<ITaskAdapter<?>>)(Collection<?>)sFAEntry.getSafetyFunctionAdapter()
						.getDiagnosticComponents();
		sfComps.stream().forEach(t -> encoding.removeTask(t));
	}

	/** {@inheritDoc} */
	@Override
	public SafeTaskGraphEncoding decode(Genotype genotype) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public SafetyFunctionArchEncoding get() {
		return new SafetyFunctionArchEncoding(null);
	}
}
