/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.evaluator.BooleanValue;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.dsl_v2.model.expression.IExpression;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.Value;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Evaluator for {@link IExpression}s of {@link ExplorationConstraint}s. The actual evaluation
 * function resides in the evaluate method.
 * 
 * @author diewald
 */
public class MappingEvaluatorConstraint
		extends
		EvaluatorModule<Boolean, ExplorationConstraint<Boolean>, TaskMappingEncoding<ITaskAdapter<?>, TaskMappingEntry<ITaskAdapter<?>, Partition>>>
		implements
		IConstraintEvaluator<Boolean, ExplorationConstraint<Boolean>, TaskMappingEncoding<ITaskAdapter<?>, TaskMappingEntry<ITaskAdapter<?>, Partition>>> {

	/** Prototype constructor. */
	public MappingEvaluatorConstraint(Class<ExplorationConstraint<Boolean>> explorationTargetType) {
		super(explorationTargetType);
	}

	/**
	 * Constructor for execution: Defines the systemModelAdapter abstraction and the
	 * {@link ExplorationTarget} to evaluate.
	 */
	public MappingEvaluatorConstraint(ExplorationConstraint<Boolean> explorationTarget) {
		super(explorationTarget);
	}

	/**
	 * Evaluates the {@link IExpression} given in the {@link ExplorationConstraint} subject to the
	 * given {@link FailSilentTaskMappingEncoding} and {@link PartitionMappingEncoding} and returns
	 * the result. Therefore, this evaluation calls the {@link MappingParserConstraint} and provides
	 * a boolean whether the expression is satisfied.
	 */
	@Provides
	@Named("MappingEvaluatorConstraint")
	public BooleanValue evaluate(FailSilentTaskMappingEncoding mapping,
			PartitionMappingEncoding virtMapping, TaskMapper<EObject> taskMapper) {
		MappingParserConstraint parser =
				new MappingParserConstraint(mapping, virtMapping,
						explorationTarget.getExpression(), taskMapper);

		Boolean retVal = null;
		// FIXME: pass parsing exception to upper layers.
		try {
			retVal = parser.parse();
		} catch(Exception e) {
			throw new ExplorationRuntimeException("Failed to parse the expression " +
					explorationTarget.getName() + ".", e);
		}

		return new BooleanValue(retVal);
	}

	/** {@inheritDoc} */
	@Override
	public
			boolean
			validate(
					ExplorationConstraint<?> expTarget,
					TaskMappingEncoding<ITaskAdapter<?>, TaskMappingEntry<ITaskAdapter<?>, Partition>> solution,
					Value<?> value) {
		// FIXME: For porting
		if(value == null) {
			return true;
		}

		Double evalValue = value.getDouble();
		if(evalValue > 0.0) {
			return false;
		}
		return true;
	}
}
