/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.graph.display;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.swing.JFrame;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IModelElementAdapter;
import org.jgrapht.Graph;
import org.jgrapht.ext.JGraphXAdapter;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.mxgraph.layout.mxFastOrganicLayout;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.util.mxConstants;

/**
 * Used for visualizing graphs (from JGraphT). To display a graph, this class must be instantiated
 * first, and then the {@code displayGraph} method must be called.
 * 
 * Usage Example:
 * {@code JGraphTVisualizer<IDeployableComponentAdapter, ChannelAdapterWeightedEdge> taskViz =
					new JGraphTVisualizer<IDeployableComponentAdapter, ChannelAdapterWeightedEdge>(
							taskGraph.getGraph());
			taskViz.displayGraph(); }
 * 
 * @author diewald
 */
public class JGraphTVisualizer<V extends IModelElementAdapter<?>, E> {

	/**
	 * Contains the graph which was given to the Constructor, except that the vertices have been
	 * converted to Named vertices to support comprehensive visualization.
	 */
	private DefaultDirectedGraph<NamedVertex<V>, DefaultEdge> namedGraph;

	/** Constructor. */
	public JGraphTVisualizer(Graph<V, E> g) {
		namedGraph = constructNamedGraph(g);
	}

	/** Displays the graph given in the constructor. */
	public void displayGraph(String graphWindowTitle) {
		// Create an adapter for JGraphX, via an adapter.
		JGraphXAdapter<NamedVertex<V>, DefaultEdge> jgxAdapter =
				new JGraphXAdapter<NamedVertex<V>, DefaultEdge>(namedGraph);

		// Disable labels of edges.
		jgxAdapter.getStylesheet().getDefaultEdgeStyle().put(mxConstants.STYLE_NOLABEL, "1");

		// positioning via jgraphx layouts
		mxFastOrganicLayout layout = new mxFastOrganicLayout(jgxAdapter);
		// layout.setMinDistanceLimit(layout.getMinDistanceLimit() * 2);
		// The higher, the more separated layout.
		layout.setForceConstant(layout.getForceConstant() * 1.5);
		layout.execute(jgxAdapter.getDefaultParent());

		mxGraphComponent graphComponent = new mxGraphComponent(jgxAdapter);

		// Draws the window containing the graph.
		JFrame frame = new JFrame();
		frame.setTitle(graphWindowTitle);
		frame.getContentPane().add(graphComponent);
		frame.pack();
		frame.setVisible(true);
	}

	/**
	 * Converts the given graph to graph that has {@link NamedVertex}-Nodes instead of the original
	 * vertices. Displaying those is more convenient.
	 */
	private DefaultDirectedGraph<NamedVertex<V>, DefaultEdge> constructNamedGraph(
			Graph<V, E> originalGraph) {
		DefaultDirectedGraph<NamedVertex<V>, DefaultEdge> namedGraph =
				new DefaultDirectedGraph<NamedVertex<V>, DefaultEdge>(DefaultEdge.class);
		Set<E> originalEdgeSet = originalGraph.edgeSet();
		Map<V, NamedVertex<V>> relateVertexToNamedElement = new HashMap<V, NamedVertex<V>>();

		for(V currentVertex : originalGraph.vertexSet()) {
			NamedVertex<V> currentNamedVertex = new NamedVertex<V>(currentVertex);
			namedGraph.addVertex(currentNamedVertex);
			relateVertexToNamedElement.put(currentVertex, currentNamedVertex);
		}

		for(E edge : originalEdgeSet) {
			namedGraph.addEdge(relateVertexToNamedElement.get(originalGraph.getEdgeSource(edge)),
					relateVertexToNamedElement.get(originalGraph.getEdgeTarget(edge)));
		}

		return namedGraph;
	}
}
